@@ -1,5 +1,6 @@
 /*
- * Copyright (c) [2012-2014] Novell, Inc.
+ * Copyright (c) [2012-2015] Novell, Inc.
+ * Copyright (c) [2016,2018] SUSE LLC
  *
  * All Rights Reserved.
  *
@@ -20,7 +21,13 @@
  */
 
 
+#include <iostream>
+
 #include "commands.h"
+#include "utils/text.h"
+#include "snapper/AppUtil.h"
+
+using namespace std;
 
 
 #define SERVICE "org.opensuse.Snapper"
@@ -28,14 +35,14 @@
 #define INTERFACE "org.opensuse.Snapper"
 
 
-list<XConfigInfo>
+vector<XConfigInfo>
 command_list_xconfigs(DBus::Connection& conn)
 {
     DBus::MessageMethodCall call(SERVICE, OBJECT, INTERFACE, "ListConfigs");
 
     DBus::Message reply = conn.send_with_reply_and_block(call);
 
-    list<XConfigInfo> ret;
+    vector<XConfigInfo> ret;
 
     DBus::Hihi hihi(reply);
     hihi >> ret;
@@ -77,8 +84,8 @@ command_set_xconfig(DBus::Connection& conn, const string& config_name,
 
 
 void
-command_create_xconfig(DBus::Connection& conn, const string& config_name, const string& subvolume,
-		       const string& fstype, const string& template_name)
+command_create_config(DBus::Connection& conn, const string& config_name, const string& subvolume,
+		      const string& fstype, const string& template_name)
 {
     DBus::MessageMethodCall call(SERVICE, OBJECT, INTERFACE, "CreateConfig");
 
@@ -90,7 +97,7 @@ command_create_xconfig(DBus::Connection& conn, const string& config_name, const
 
 
 void
-command_delete_xconfig(DBus::Connection& conn, const string& config_name)
+command_delete_config(DBus::Connection& conn, const string& config_name)
 {
     DBus::MessageMethodCall call(SERVICE, OBJECT, INTERFACE, "DeleteConfig");
 
@@ -140,22 +147,22 @@ command_get_xsnapshot(DBus::Connection& conn, const string& config_name, unsigne
 
 
 void
-command_set_xsnapshot(DBus::Connection& conn, const string& config_name, unsigned int num,
-		      const XSnapshot& data)
+command_set_snapshot(DBus::Connection& conn, const string& config_name, unsigned int num,
+		     const SMD& smd)
 {
     DBus::MessageMethodCall call(SERVICE, OBJECT, INTERFACE, "SetSnapshot");
 
     DBus::Hoho hoho(call);
-    hoho << config_name << num << data.description << data.cleanup << data.userdata;
+    hoho << config_name << num << smd.description << smd.cleanup << smd.userdata;
 
     conn.send_with_reply_and_block(call);
 }
 
 
 unsigned int
-command_create_single_xsnapshot(DBus::Connection& conn, const string& config_name,
-				const string& description, const string& cleanup,
-				const map<string, string>& userdata)
+command_create_single_snapshot(DBus::Connection& conn, const string& config_name,
+			       const string& description, const string& cleanup,
+			       const map<string, string>& userdata)
 {
     DBus::MessageMethodCall call(SERVICE, OBJECT, INTERFACE, "CreateSingleSnapshot");
 
@@ -174,10 +181,10 @@ command_create_single_xsnapshot(DBus::Connection& conn, const string& config_nam
 
 
 unsigned int
-command_create_single_xsnapshot_v2(DBus::Connection& conn, const string& config_name,
-				   unsigned int parent_num, bool read_only,
-				   const string& description, const string& cleanup,
-				   const map<string, string>& userdata)
+command_create_single_snapshot_v2(DBus::Connection& conn, const string& config_name,
+				  unsigned int parent_num, bool read_only,
+				  const string& description, const string& cleanup,
+				  const map<string, string>& userdata)
 {
     DBus::MessageMethodCall call(SERVICE, OBJECT, INTERFACE, "CreateSingleSnapshotV2");
 
@@ -196,10 +203,10 @@ command_create_single_xsnapshot_v2(DBus::Connection& conn, const string& config_
 
 
 unsigned int
-command_create_single_xsnapshot_of_default(DBus::Connection& conn, const string& config_name,
-					   bool read_only, const string& description,
-					   const string& cleanup,
-					   const map<string, string>& userdata)
+command_create_single_snapshot_of_default(DBus::Connection& conn, const string& config_name,
+					  bool read_only, const string& description,
+					  const string& cleanup,
+					  const map<string, string>& userdata)
 {
     DBus::MessageMethodCall call(SERVICE, OBJECT, INTERFACE, "CreateSingleSnapshotOfDefault");
 
@@ -218,9 +225,9 @@ command_create_single_xsnapshot_of_default(DBus::Connection& conn, const string&
 
 
 unsigned int
-command_create_pre_xsnapshot(DBus::Connection& conn, const string& config_name,
-			     const string& description, const string& cleanup,
-			     const map<string, string>& userdata)
+command_create_pre_snapshot(DBus::Connection& conn, const string& config_name,
+			    const string& description, const string& cleanup,
+			    const map<string, string>& userdata)
 {
     DBus::MessageMethodCall call(SERVICE, OBJECT, INTERFACE, "CreatePreSnapshot");
 
@@ -239,9 +246,9 @@ command_create_pre_xsnapshot(DBus::Connection& conn, const string& config_name,
 
 
 unsigned int
-command_create_post_xsnapshot(DBus::Connection& conn, const string& config_name,
-			      unsigned int prenum, const string& description,
-			      const string& cleanup, const map<string, string>& userdata)
+command_create_post_snapshot(DBus::Connection& conn, const string& config_name,
+			     unsigned int prenum, const string& description,
+			     const string& cleanup, const map<string, string>& userdata)
 {
     DBus::MessageMethodCall call(SERVICE, OBJECT, INTERFACE, "CreatePostSnapshot");
 
@@ -260,9 +267,23 @@ command_create_post_xsnapshot(DBus::Connection& conn, const string& config_name,
 
 
 void
-command_delete_xsnapshots(DBus::Connection& conn, const string& config_name,
-			  list<unsigned int> nums)
+command_delete_snapshots(DBus::Connection& conn, const string& config_name,
+			 const vector<unsigned int>& nums, bool verbose)
 {
+    if (verbose)
+    {
+	cout << sformat(_("Deleting snapshot from %s:", "Deleting snapshots from %s:", nums.size()),
+			config_name.c_str()) << endl;
+
+	for (vector<unsigned int>::const_iterator it = nums.begin(); it != nums.end(); ++it)
+	{
+	    if (it != nums.begin())
+		cout << ", ";
+	    cout << *it;
+	}
+	cout << endl;
+    }
+
     DBus::MessageMethodCall call(SERVICE, OBJECT, INTERFACE, "DeleteSnapshots");
 
     DBus::Hoho hoho(call);
@@ -272,9 +293,52 @@ command_delete_xsnapshots(DBus::Connection& conn, const string& config_name,
 }
 
 
+void
+command_calculate_used_space(DBus::Connection& conn, const string& config_name)
+{
+    DBus::MessageMethodCall call(SERVICE, OBJECT, INTERFACE, "CalculateUsedSpace");
+
+    DBus::Hoho hoho(call);
+    hoho << config_name;
+
+    try
+    {
+	conn.send_with_reply_and_block(call);
+    }
+    catch (const DBus::ErrorException& e)
+    {
+	SN_CAUGHT(e);
+
+	if (strcmp(e.name(), "error.quota") == 0)
+	    SN_THROW(QuotaException(e.message()));
+
+	SN_RETHROW(e);
+    }
+}
+
+
+uint64_t
+command_get_used_space(DBus::Connection& conn, const string& config_name, unsigned int num)
+{
+    DBus::MessageMethodCall call(SERVICE, OBJECT, INTERFACE, "GetUsedSpace");
+
+    DBus::Hoho hoho(call);
+    hoho << config_name << num;
+
+    DBus::Message reply = conn.send_with_reply_and_block(call);
+
+    uint64_t used_space;
+
+    DBus::Hihi hihi(reply);
+    hihi >> used_space;
+
+    return used_space;
+}
+
+
 string
-command_mount_xsnapshots(DBus::Connection& conn, const string& config_name,
-			 unsigned int num, bool user_request)
+command_mount_snapshot(DBus::Connection& conn, const string& config_name,
+		       unsigned int num, bool user_request)
 {
     DBus::MessageMethodCall call(SERVICE, OBJECT, INTERFACE, "MountSnapshot");
 
@@ -293,8 +357,8 @@ command_mount_xsnapshots(DBus::Connection& conn, const string& config_name,
 
 
 void
-command_umount_xsnapshots(DBus::Connection& conn, const string& config_name,
-			  unsigned int num, bool user_request)
+command_umount_snapshot(DBus::Connection& conn, const string& config_name, unsigned int num,
+			bool user_request)
 {
     DBus::MessageMethodCall call(SERVICE, OBJECT, INTERFACE, "UmountSnapshot");
 
@@ -306,8 +370,7 @@ command_umount_xsnapshots(DBus::Connection& conn, const string& config_name,
 
 
 string
-command_get_xmount_point(DBus::Connection& conn, const string& config_name,
-			 unsigned int num)
+command_get_mount_point(DBus::Connection& conn, const string& config_name, unsigned int num)
 {
     DBus::MessageMethodCall call(SERVICE, OBJECT, INTERFACE, "GetMountPoint");
 
@@ -326,8 +389,8 @@ command_get_xmount_point(DBus::Connection& conn, const string& config_name,
 
 
 void
-command_create_xcomparison(DBus::Connection& conn, const string& config_name, unsigned int number1,
-			   unsigned int number2)
+command_create_comparison(DBus::Connection& conn, const string& config_name, unsigned int number1,
+			  unsigned int number2)
 {
     DBus::MessageMethodCall call(SERVICE, OBJECT, INTERFACE, "CreateComparison");
 
@@ -339,8 +402,8 @@ command_create_xcomparison(DBus::Connection& conn, const string& config_name, un
 
 
 void
-command_delete_xcomparison(DBus::Connection& conn, const string& config_name, unsigned int number1,
-			   unsigned int number2)
+command_delete_comparison(DBus::Connection& conn, const string& config_name, unsigned int number1,
+			  unsigned int number2)
 {
     DBus::MessageMethodCall call(SERVICE, OBJECT, INTERFACE, "DeleteComparison");
 
@@ -351,7 +414,14 @@ command_delete_xcomparison(DBus::Connection& conn, const string& config_name, un
 }
 
 
-list<XFile>
+int
+operator<(const XFile& lhs, const XFile& rhs)
+{
+    return File::cmp_lt(lhs.name, rhs.name);
+}
+
+
+vector<XFile>
 command_get_xfiles(DBus::Connection& conn, const string& config_name, unsigned int number1,
 		   unsigned int number2)
 {
@@ -362,17 +432,75 @@ command_get_xfiles(DBus::Connection& conn, const string& config_name, unsigned i
 
     DBus::Message reply = conn.send_with_reply_and_block(call);
 
-    list<XFile> files;
+    vector<XFile> files;
 
     DBus::Hihi hihi(reply);
     hihi >> files;
 
+    sort(files.begin(), files.end());	// snapperd can have different locale than client
+					// so sorting is required here
+
     return files;
 }
 
 
+void
+command_setup_quota(DBus::Connection& conn, const string& config_name)
+{
+    DBus::MessageMethodCall call(SERVICE, OBJECT, INTERFACE, "SetupQuota");
+
+    DBus::Hoho hoho(call);
+    hoho << config_name;
+
+    conn.send_with_reply_and_block(call);
+}
+
+
+void
+command_prepare_quota(DBus::Connection& conn, const string& config_name)
+{
+    DBus::MessageMethodCall call(SERVICE, OBJECT, INTERFACE, "PrepareQuota");
+
+    DBus::Hoho hoho(call);
+    hoho << config_name;
+
+    conn.send_with_reply_and_block(call);
+}
+
+
+QuotaData
+command_query_quota(DBus::Connection& conn, const string& config_name)
+{
+    DBus::MessageMethodCall call(SERVICE, OBJECT, INTERFACE, "QueryQuota");
+
+    DBus::Hoho hoho(call);
+    hoho << config_name;
+
+    DBus::Message reply = conn.send_with_reply_and_block(call);
+
+    QuotaData quota_data;
+
+    DBus::Hihi hihi(reply);
+    hihi >> quota_data;
+
+    return quota_data;
+}
+
+
+void
+command_sync(DBus::Connection& conn, const string& config_name)
+{
+    DBus::MessageMethodCall call(SERVICE, OBJECT, INTERFACE, "Sync");
+
+    DBus::Hoho hoho(call);
+    hoho << config_name;
+
+    conn.send_with_reply_and_block(call);
+}
+
+
 vector<string>
-command_xdebug(DBus::Connection& conn)
+command_debug(DBus::Connection& conn)
 {
     DBus::MessageMethodCall call(SERVICE, OBJECT, INTERFACE, "Debug");
 
