@@ -1,5 +1,6 @@
 /*
  * Copyright (c) [2011-2015] Novell, Inc.
+ * Copyright (c) [2016,2018] SUSE LLC
  *
  * All Rights Reserved.
  *
@@ -24,6 +25,7 @@
 
 #include <sys/stat.h>
 #include <sys/types.h>
+#include <sys/statvfs.h>
 #include <glob.h>
 #include <string.h>
 #include <mntent.h>
@@ -44,6 +46,12 @@
 #include "snapper/AsciiFile.h"
 #include "snapper/Exception.h"
 #include "snapper/Hooks.h"
+#include "snapper/Btrfs.h"
+#include "snapper/BtrfsUtils.h"
+#ifdef ENABLE_SELINUX
+#include "snapper/Selinux.h"
+#include "snapper/Regex.h"
+#endif
 
 
 namespace snapper
@@ -56,7 +64,7 @@ namespace snapper
 	  config_name(config_name), subvolume("/")
     {
 	if (!getValue(KEY_SUBVOLUME, subvolume))
-	    throw InvalidConfigException();
+	    SN_THROW(InvalidConfigException());
     }
 
 
@@ -64,7 +72,7 @@ namespace snapper
     ConfigInfo::checkKey(const string& key) const
     {
 	if (key == KEY_SUBVOLUME || key == KEY_FSTYPE)
-	    throw InvalidConfigdataException();
+	    SN_THROW(InvalidConfigdataException());
 
 	try
 	{
@@ -72,29 +80,43 @@ namespace snapper
 	}
 	catch (const InvalidKeyException& e)
 	{
-	    throw InvalidConfigdataException();
+	    SN_THROW(InvalidConfigdataException());
 	}
     }
 
 
     Snapper::Snapper(const string& config_name, const string& root_prefix, bool disable_filters)
-	: config_info(NULL), filesystem(NULL), snapshots(this)
+	: config_info(NULL), filesystem(NULL), snapshots(this), selabel_handle(NULL)
     {
 	y2mil("Snapper constructor");
 	y2mil("libsnapper version " VERSION);
 	y2mil("config_name:" << config_name << " disable_filters:" << disable_filters);
 
+#ifdef ENABLE_SELINUX
+	try
+	{
+	    selabel_handle = SelinuxLabelHandle::get_selinux_handle();
+	}
+	catch (const SelinuxException& e)
+	{
+	    SN_RETHROW(e);
+	}
+#endif
+
 	try
 	{
 	    config_info = new ConfigInfo(config_name, root_prefix);
 	}
 	catch (const FileNotFoundException& e)
 	{
-	    throw ConfigNotFoundException();
+	    SN_THROW(ConfigNotFoundException());
 	}
 
 	filesystem = Filesystem::create(*config_info, root_prefix);
 
+	// With btrfs backend, it's useless try syncing snapshot RO subvolumes
+	syncSelinuxContexts(filesystem->fstype() == "btrfs");
+
 	bool sync_acl;
 	if (config_info->getValue(KEY_SYNC_ACL, sync_acl) && sync_acl == true)
 	    syncAcl();
@@ -125,7 +147,10 @@ namespace snapper
 	}
 
 	delete filesystem;
+	filesystem = nullptr;
+
 	delete config_info;
+	config_info = nullptr;
     }
 
 
@@ -193,7 +218,7 @@ namespace snapper
     Snapper::createSingleSnapshot(Snapshots::const_iterator parent, const SCD& scd)
     {
 	if (parent == snapshots.end())
-	    throw IllegalSnapshotException();
+	    SN_THROW(IllegalSnapshotException());
 
 	return snapshots.createSingleSnapshot(parent, scd);
     }
@@ -251,7 +276,7 @@ namespace snapper
 
 	try
 	{
-	    SysconfigFile sysconfig(SYSCONFIGFILE);
+	    SysconfigFile sysconfig(prepend_root_prefix(root_prefix, SYSCONFIGFILE));
 	    vector<string> config_names;
 	    sysconfig.getValue("SNAPPER_CONFIGS", config_names);
 
@@ -273,7 +298,7 @@ namespace snapper
 	}
 	catch (const FileNotFoundException& e)
 	{
-	    throw ListConfigsFailedException("sysconfig-file not found");
+	    SN_THROW(ListConfigsFailedException("sysconfig-file not found"));
 	}
 
 	return config_infos;
@@ -292,12 +317,12 @@ namespace snapper
 
 	if (config_name.empty() || config_name.find_first_of(", \t") != string::npos)
 	{
-	    throw CreateConfigFailedException("illegal config name");
+	    SN_THROW(CreateConfigFailedException("illegal config name"));
 	}
 
 	if (!boost::starts_with(subvolume, "/") || !checkDir(subvolume))
 	{
-	    throw CreateConfigFailedException("illegal subvolume");
+	    SN_THROW(CreateConfigFailedException("illegal subvolume"));
 	}
 
 	list<ConfigInfo> configs = getConfigs(root_prefix);
@@ -305,13 +330,13 @@ namespace snapper
 	{
 	    if (it->getSubvolume() == subvolume)
 	    {
-		throw CreateConfigFailedException("subvolume already covered");
+		SN_THROW(CreateConfigFailedException("subvolume already covered"));
 	    }
 	}
 
 	if (access(string(CONFIGTEMPLATEDIR "/" + template_name).c_str(), R_OK) != 0)
 	{
-	    throw CreateConfigFailedException("cannot access template config");
+	    SN_THROW(CreateConfigFailedException("cannot access template config"));
 	}
 
 	unique_ptr<Filesystem> filesystem;
@@ -321,11 +346,11 @@ namespace snapper
 	}
 	catch (const InvalidConfigException& e)
 	{
-	    throw CreateConfigFailedException("invalid filesystem type");
+	    SN_THROW(CreateConfigFailedException("invalid filesystem type"));
 	}
 	catch (const ProgramNotInstalledException& e)
 	{
-	    throw CreateConfigFailedException(e.what());
+	    SN_THROW(CreateConfigFailedException(e.what()));
 	}
 
 	try
@@ -335,7 +360,7 @@ namespace snapper
 	    sysconfig.getValue("SNAPPER_CONFIGS", config_names);
 	    if (find(config_names.begin(), config_names.end(), config_name) != config_names.end())
 	    {
-		throw CreateConfigFailedException("config already exists");
+		SN_THROW(CreateConfigFailedException("config already exists"));
 	    }
 
 	    config_names.push_back(config_name);
@@ -343,7 +368,7 @@ namespace snapper
 	}
 	catch (const FileNotFoundException& e)
 	{
-	    throw CreateConfigFailedException("sysconfig-file not found");
+	    SN_THROW(CreateConfigFailedException("sysconfig-file not found"));
 	}
 
 	try
@@ -357,15 +382,17 @@ namespace snapper
 	}
 	catch (const FileNotFoundException& e)
 	{
-	    throw CreateConfigFailedException("modifying config failed");
+	    SN_THROW(CreateConfigFailedException("modifying config failed"));
 	}
 
 	try
 	{
 	    filesystem->createConfig();
 	}
-	catch (...)
+	catch (const Exception& e)
 	{
+	    SN_CAUGHT(e);
+
 	    SysconfigFile sysconfig(SYSCONFIGFILE);
 	    vector<string> config_names;
 	    sysconfig.getValue("SNAPPER_CONFIGS", config_names);
@@ -375,7 +402,7 @@ namespace snapper
 
 	    SystemCmd cmd(RMBIN " " + quote(CONFIGSDIR "/" + config_name));
 
-	    throw;
+	    SN_RETHROW(e);
 	}
 
 	Hooks::create_config(subvolume, filesystem.get());
@@ -416,13 +443,13 @@ namespace snapper
 	}
 	catch (const DeleteConfigFailedException& e)
 	{
-	    throw DeleteConfigFailedException("deleting snapshot failed");
+	    SN_THROW(DeleteConfigFailedException("deleting snapshot failed"));
 	}
 
 	SystemCmd cmd1(RMBIN " " + quote(CONFIGSDIR "/" + config_name));
 	if (cmd1.retcode() != 0)
 	{
-	    throw DeleteConfigFailedException("deleting config-file failed");
+	    SN_THROW(DeleteConfigFailedException("deleting config-file failed"));
 	}
 
 	try
@@ -436,7 +463,7 @@ namespace snapper
 	}
 	catch (const FileNotFoundException& e)
 	{
-	    throw DeleteConfigFailedException("sysconfig-file not found");
+	    SN_THROW(DeleteConfigFailedException("sysconfig-file not found"));
 	}
     }
 
@@ -449,6 +476,8 @@ namespace snapper
 
 	config_info->save();
 
+	filesystem->evalConfigInfo(*config_info);
+
 	if (raw.find(KEY_ALLOW_USERS) != raw.end() || raw.find(KEY_ALLOW_GROUPS) != raw.end() ||
 	    raw.find(KEY_SYNC_ACL) != raw.end())
 	{
@@ -470,7 +499,7 @@ namespace snapper
 	    {
 		uid_t uid;
 		if (!get_user_uid(it->c_str(), uid))
-		    throw InvalidUserException();
+		    SN_THROW(InvalidUserException());
 		uids.push_back(uid);
 	    }
 	}
@@ -483,7 +512,7 @@ namespace snapper
 	    {
 		gid_t gid;
 		if (!get_group_gid(it->c_str(), gid))
-		    throw InvalidGroupException();
+		    SN_THROW(InvalidGroupException());
 		gids.push_back(gid);
 	    }
 	}
@@ -492,16 +521,23 @@ namespace snapper
     }
 
 
+    void
+    Snapper::syncFilesystem() const
+    {
+	filesystem->sync();
+    }
+
+
     static void
     set_acl_permissions(acl_entry_t entry)
     {
 	acl_permset_t permset;
 	if (acl_get_permset(entry, &permset) != 0)
-	    throw AclException();
+	    SN_THROW(AclException());
 
 	if (acl_add_perm(permset, ACL_READ) != 0 || acl_delete_perm(permset, ACL_WRITE) != 0 ||
 	    acl_add_perm(permset, ACL_EXECUTE) != 0)
-	    throw AclException();
+	    SN_THROW(AclException());
     };
 
 
@@ -510,13 +546,13 @@ namespace snapper
     {
 	acl_entry_t entry;
 	if (acl_create_entry(acl, &entry) != 0)
-	    throw AclException();
+	    SN_THROW(AclException());
 
 	if (acl_set_tag_type(entry, tag) != 0)
-	    throw AclException();
+	    SN_THROW(AclException());
 
 	if (acl_set_qualifier(entry, qualifier) != 0)
-	    throw AclException();
+	    SN_THROW(AclException());
 
 	set_acl_permissions(entry);
     };
@@ -529,11 +565,11 @@ namespace snapper
 
 	acl_t orig_acl = acl_get_fd(infos_dir.fd());
 	if (!orig_acl)
-	    throw AclException();
+	    SN_THROW(AclException());
 
 	acl_t acl = acl_dup(orig_acl);
 	if (!acl)
-	    throw AclException();
+	    SN_THROW(AclException());
 
 	set<uid_t> remaining_uids = set<uid_t>(uids.begin(), uids.end());
 	set<gid_t> remaining_gids = set<gid_t>(gids.begin(), gids.end());
@@ -545,7 +581,7 @@ namespace snapper
 
 		acl_tag_t tag;
 		if (acl_get_tag_type(entry, &tag) != 0)
-		    throw AclException();
+		    SN_THROW(AclException());
 
 		switch (tag)
 		{
@@ -553,7 +589,7 @@ namespace snapper
 
 			uid_t* uid = (uid_t*) acl_get_qualifier(entry);
 			if (!uid)
-			    throw AclException();
+			    SN_THROW(AclException());
 
 			if (contains(remaining_uids, *uid))
 			{
@@ -564,7 +600,7 @@ namespace snapper
 			else
 			{
 			    if (acl_delete_entry(acl, entry) != 0)
-				throw AclException();
+				SN_THROW(AclException());
 			}
 
 		    } break;
@@ -573,7 +609,7 @@ namespace snapper
 
 			gid_t* gid = (gid_t*) acl_get_qualifier(entry);
 			if (!gid)
-			    throw AclException();
+			    SN_THROW(AclException());
 
 			if (contains(remaining_gids, *gid))
 			{
@@ -584,7 +620,7 @@ namespace snapper
 			else
 			{
 			    if (acl_delete_entry(acl, entry) != 0)
-				throw AclException();
+				SN_THROW(AclException());
 			}
 
 		    } break;
@@ -605,14 +641,253 @@ namespace snapper
 	}
 
 	if (acl_calc_mask(&acl) != 0)
-	    throw AclException();
+	    SN_THROW(AclException());
 
 	if (acl_cmp(orig_acl, acl) == 1)
 	    if (acl_set_fd(infos_dir.fd(), acl) != 0)
-		throw AclException();
+		SN_THROW(AclException());
 
 	if (acl_free(acl) != 0)
-	    throw AclException();
+	    SN_THROW(AclException());
+    }
+
+
+    void
+    Snapper::setupQuota()
+    {
+#ifdef ENABLE_BTRFS_QUOTA
+
+	const Btrfs* btrfs = dynamic_cast<const Btrfs*>(getFilesystem());
+	if (!btrfs)
+	    SN_THROW(QuotaException("quota only supported with btrfs"));
+
+	if (btrfs->getQGroup() != no_qgroup)
+	    SN_THROW(QuotaException("qgroup already set"));
+
+	SDir general_dir = btrfs->openGeneralDir();
+
+	quota_enable(general_dir.fd());
+
+	qgroup_t qgroup = qgroup_find_free(general_dir.fd(), 1);
+
+	y2mil("free qgroup:" << format_qgroup(qgroup));
+
+	qgroup_create(general_dir.fd(), qgroup);
+
+	setConfigInfo({ { "QGROUP", format_qgroup(qgroup) } });
+
+#else
+
+	SN_THROW(QuotaException("not implemented"));
+	__builtin_unreachable();
+
+#endif
+    }
+
+
+    void
+    Snapper::prepareQuota() const
+    {
+#ifdef ENABLE_BTRFS_QUOTA
+
+	const Btrfs* btrfs = dynamic_cast<const Btrfs*>(getFilesystem());
+	if (!btrfs)
+	    SN_THROW(QuotaException("quota only supported with btrfs"));
+
+	if (btrfs->getQGroup() == no_qgroup)
+	    SN_THROW(QuotaException("qgroup not set"));
+
+	SDir general_dir = btrfs->openGeneralDir();
+
+	try
+	{
+	    vector<qgroup_t> children = qgroup_query_children(general_dir.fd(), btrfs->getQGroup());
+	    sort(children.begin(), children.end());
+
+	    // Iterate all snapshots and ensure that those and only those with
+	    // a cleanup algorithm are included in the high level qgroup.
+
+	    for (const Snapshot& snapshot : snapshots)
+	    {
+		if (snapshot.isCurrent())
+		    continue;
+
+		subvolid_t subvolid = get_id(snapshot.openSnapshotDir().fd());
+		qgroup_t qgroup = calc_qgroup(0, subvolid);
+
+		bool included = binary_search(children.begin(), children.end(), qgroup);
+
+		if (!snapshot.getCleanup().empty() && !included)
+		{
+		    qgroup_assign(general_dir.fd(), qgroup, btrfs->getQGroup());
+		}
+		else if (snapshot.getCleanup().empty() && included)
+		{
+		    qgroup_remove(general_dir.fd(), qgroup, btrfs->getQGroup());
+		}
+	    }
+	}
+	catch (const runtime_error& e)
+	{
+	    SN_THROW(QuotaException("preparing quota failed"));
+	}
+
+#else
+
+	SN_THROW(QuotaException("not implemented"));
+	__builtin_unreachable();
+
+#endif
+    }
+
+
+    void
+    Snapper::calculateUsedSpace() const
+    {
+#ifdef ENABLE_BTRFS_QUOTA
+
+	const Btrfs* btrfs = dynamic_cast<const Btrfs*>(getFilesystem());
+	if (!btrfs)
+	    SN_THROW(QuotaException("quota only supported with btrfs"));
+
+	try
+	{
+	    SDir general_dir = btrfs->openGeneralDir();
+
+	    quota_rescan(general_dir.fd());
+	    sync(general_dir.fd());
+	}
+	catch (...)
+	{
+	    SN_THROW(QuotaException("quota rescan or sync failed"));
+	}
+
+#else
+
+	SN_THROW(QuotaException("not implemented"));
+	__builtin_unreachable();
+
+#endif
+    }
+
+
+    QuotaData
+    Snapper::queryQuotaData() const
+    {
+#ifdef ENABLE_BTRFS_QUOTA
+
+	const Btrfs* btrfs = dynamic_cast<const Btrfs*>(getFilesystem());
+	if (!btrfs)
+	    SN_THROW(QuotaException("quota only supported with btrfs"));
+
+	if (btrfs->getQGroup() == no_qgroup)
+	    SN_THROW(QuotaException("qgroup not set"));
+
+	SDir general_dir = btrfs->openGeneralDir();
+
+	// Tests have shown that without a rescan and sync here the quota data
+	// is incorrect.
+
+	quota_rescan(general_dir.fd());
+	sync(general_dir.fd());
+
+	struct statvfs64 fsbuf;
+	if (fstatvfs64(general_dir.fd(), &fsbuf) != 0)
+	    SN_THROW(QuotaException("statvfs64 failed"));
+
+	QuotaData quota_data;
+
+	quota_data.size = fsbuf.f_blocks * fsbuf.f_bsize;
+
+	QGroupUsage qgroup_usage = qgroup_query_usage(general_dir.fd(), btrfs->getQGroup());
+	quota_data.used = qgroup_usage.exclusive;
+
+	y2mil("size:" << quota_data.size << " used:" << quota_data.used);
+
+	if (quota_data.used > quota_data.size)
+	    SN_THROW(QuotaException("impossible quota values"));
+
+	return quota_data;
+
+#else
+
+	SN_THROW(QuotaException("not implemented"));
+	__builtin_unreachable();
+
+#endif
+    }
+
+
+    void
+    Snapper::syncSelinuxContexts(bool skip_snapshot_dir) const
+    {
+#ifdef ENABLE_SELINUX
+	try
+	{
+	    SDir subvol_dir = openSubvolumeDir();
+	    SDir infos_dir(subvol_dir, ".snapshots");
+
+	    if (infos_dir.restorecon(selabel_handle))
+	    {
+		syncSelinuxContextsInInfosDir(skip_snapshot_dir);
+	    }
+	    else
+	    {
+		SnapperContexts scons;
+
+		if (infos_dir.fsetfilecon(scons.subvolume_context()))
+		    syncSelinuxContextsInInfosDir(skip_snapshot_dir);
+	    }
+	}
+	catch (const SelinuxException& e)
+	{
+	    SN_CAUGHT(e);
+	    // fall through intentional
+	}
+#endif
+    }
+
+
+    void
+    Snapper::syncSelinuxContextsInInfosDir(bool skip_snapshot_dir) const
+    {
+#ifdef ENABLE_SELINUX
+	Regex rx("^[0-9]+$");
+	Regex rx_filelist("^filelist-[0-9]+.txt$");
+
+	y2deb("Syncing Selinux contexts in infos dir");
+
+	SDir infos_dir = openInfosDir();
+
+	vector<string> infos = infos_dir.entries();
+	for (vector<string>::const_iterator it1 = infos.begin(); it1 != infos.end(); ++it1)
+	{
+	    if (!rx.match(*it1))
+		continue;
+
+	    SDir info_dir(infos_dir, *it1);
+	    info_dir.restorecon(selabel_handle);
+
+	    SFile info(info_dir, "info.xml");
+	    info.restorecon(selabel_handle);
+
+	    if (!skip_snapshot_dir)
+	    {
+		SFile snapshot_dir(info_dir, "snapshot");
+		snapshot_dir.restorecon(selabel_handle);
+	    }
+
+	    vector<string> info_content = info_dir.entries();
+	    for (vector<string>::const_iterator it2 = info_content.begin(); it2 != info_content.end(); ++it2)
+	    {
+		if (!rx_filelist.match(*it2))
+		    continue;
+
+		SFile fl(info_dir, *it2);
+		fl.restorecon(selabel_handle);
+	    }
+	}
+#endif
     }
 
 
@@ -714,9 +989,22 @@ namespace snapper
 #ifndef ENABLE_BTRFS_QUOTA
 	    "no-"
 #endif
-	    "btrfs-quota"
+	    "btrfs-quota,"
+
+#ifndef ENABLE_SELINUX
+	    "no-"
+#endif
+	    "selinux"
 
 	    ;
     }
 
+
+    vector<string>
+    Snapper::debug()
+    {
+	return { "version " + string(compileVersion()),
+		 "flags " + string(compileFlags()) };
+    }
+
 }
